"""
This is the module that contains the main transformation from ufl to loopy
(with pdelab as the hardcoded generation target)
"""

from __future__ import absolute_import

from dune.perftool import Restriction
from dune.perftool.ufl.modified_terminals import ModifiedTerminalTracker
from dune.perftool.pymbolic.uflmapper import UFL2PymbolicMapper
from dune.perftool.pdelab.geometry import GeometryMapper
from dune.perftool.generation import (domain,
                                      get_temporary_name,
                                      global_context,
                                      globalarg,
                                      iname,
                                      instruction,
                                      symbol,
                                      temporary_variable,
                                      valuearg,
                                      )

from dune.perftool.pdelab.basis import (lfs_iname,
                                        name_leaf_lfs,
                                        name_lfs,
                                        name_lfs_bound,
                                        traverse_lfs_tree,
                                        )
from dune.perftool.pdelab.quadrature import quadrature_iname
from pymbolic.primitives import Subscript, Variable


class UFL2LoopyVisitor(ModifiedTerminalTracker, UFL2PymbolicMapper, GeometryMapper):
    def __init__(self, measure, subdomain_id, dimension_indices):
        # Some variables describing the integral measure of this integral
        self.measure = measure
        self.subdomain_id = subdomain_id
        self.dimension_indices = dimension_indices

        # Call base class constructors
        super(UFL2LoopyVisitor, self).__init__()

    def __call__(self, o):
        # Reset some state variables that are reinitialized for each accumulation term
        self.argshape = 0
        self.redinames = ()
        self.inames = []
        self.substitution_rules = []

        # Initialize the local function spaces that we might need for this term
        # We therefore gather a list of modified trial functions too.
        from dune.perftool.ufl.modified_terminals import extract_modified_arguments
        test_ma = extract_modified_arguments(o,)
        trial_ma = extract_modified_arguments(o, coeffcount=0)
        apply_ma = extract_modified_arguments(o, coeffcount=1)

        # All test and trial functions on boundary integrals are technically restricted
        import itertools
        for ma in itertools.chain(test_ma, trial_ma, apply_ma):
            if self.measure == 'exterior_facet':
                ma.restriction = Restriction.NEGATIVE

            traverse_lfs_tree(ma)

        # Determine the rank of the term
        self.rank = len(test_ma)

        # First we do the tree traversal to get a pymbolic expression representing this expression
        pymbolic_expr = self.call(o)

        # It may happen that an entire accumulation term vanishes. We do nothing in that case
        if pymbolic_expr == 0:
            return

        # Collect the arguments for the accumulate function
        accumargs = [None] * (2 * len(test_ma))
        residual_shape = [None] * len(test_ma)
        arg_restr = [None] * len(test_ma)

        for ma in test_ma:
            count = ma.argexpr.number()
            if len(test_ma) == 2:
                icount = (count + 1) % 2
            else:
                icount = 0

            # Extract the subelement of the (mixed) finite element
            from ufl.functionview import select_subelement
            subel = select_subelement(ma.argexpr.ufl_element(), ma.component)

            # And generate a local function space for it!
            lfs = name_lfs(ma.argexpr.ufl_element(), ma.restriction, ma.component)
            from dune.perftool.generation import valuearg
            from loopy.types import NumpyType
            valuearg(lfs, dtype=NumpyType("str"))

            if len(subel.value_shape()) != 0:
                from dune.perftool.pdelab.geometry import dimension_iname
                from dune.perftool.pdelab.basis import lfs_child
                idims = tuple(dimension_iname(context='arg', count=i) for i in range(len(subel.value_shape())))
                lfs = lfs_child(lfs, idims, shape=subel.value_shape())
                residual_shape[icount] = name_lfs_bound(name_leaf_lfs(subel.sub_elements()[0], ma.restriction))
                subel = subel.sub_elements()[0]
            else:
                residual_shape[icount] = name_lfs_bound(lfs)

            lfsi = lfs_iname(subel, ma.restriction, count=count)

            # If the LFS is not yet a pymbolic expression, make it one
            from pymbolic.primitives import Expression
            if not isinstance(lfs, Expression):
                lfs = Variable(lfs)

            accumargs[2 * icount] = lfs
            accumargs[2 * icount + 1] = Variable(lfsi)

            arg_restr[icount] = ma.restriction

        from dune.perftool.pdelab.argument import name_accumulation_variable
        accumvar = name_accumulation_variable(arg_restr)

        from dune.perftool.pdelab.quadrature import name_factor
        factor = name_factor()

        predicates = frozenset({})

        # Maybe wrap this instruction into a condiditional. This mostly happens with mixed boundary conditions
        if self.subdomain_id not in ['everywhere', 'otherwise']:
            # We need to reconstruct the subdomain_data parameter of the measure
            # I am *totally* confused as to why this information is not at hand anyway,
            # but conversation with Martin pointed me to dolfin.fem.assembly where this
            # is done in preprocessing with the limitation of only one possible type of
            # modified measure per integral type.

            # Get the original form and inspect the present measures
            from dune.perftool.generation import get_global_context_value
            original_form = get_global_context_value("formdata").original_form

            sd = original_form.subdomain_data()
            assert len(sd) == 1
            subdomains, = list(sd.values())
            domain, = list(sd.keys())
            for k in list(subdomains.keys()):
                if subdomains[k] is None:
                    del subdomains[k]

            # Finally extract the original subdomain_data (which needs to be present!)
            assert self.measure in subdomains
            subdomain_data = subdomains[self.measure]

            # Determine the name of the parameter function
            name = get_global_context_value("data").object_names[id(subdomain_data)]

            # Trigger the generation of code for this thing in the parameter class
            from ufl.checks import is_cellwise_constant
            cellwise_constant = is_cellwise_constant(o)
            from dune.perftool.pdelab.parameter import intersection_parameter_function
            intersection_parameter_function(name, subdomain_data, cellwise_constant, t='int')

            predicates = predicates.union(['{} == {}'.format(name, self.subdomain_id)])

        from dune.perftool.loopy.functions import PDELabAccumulationFunction
        from pymbolic.primitives import Call, Product
        expr = Product((pymbolic_expr, Variable(factor)))
        expr = Call(PDELabAccumulationFunction(accumvar, len(test_ma)), tuple(a for a in accumargs) + (expr,))

        instruction(assignees=frozenset({}),
                    expression=expr,
                    forced_iname_deps=frozenset(self.inames).union(frozenset({quadrature_iname()})),
                    forced_iname_deps_is_final=True,
                    predicates=predicates
                    )

    def argument(self, o):
        # Correct the restriction on boundary integrals
        restriction = self.restriction
        if self.measure == 'exterior_facet':
            restriction = Restriction.NEGATIVE

        # Select the correct subtree of the finite element
        from ufl.functionview import select_subelement
        element = select_subelement(o.ufl_element(), self.component)
        leaf_element = element

        # Now treat the case of this being a vector finite element
        if element.num_sub_elements() > 0:
            # I cannot handle general mixed elements here...
            from ufl import VectorElement, TensorElement
            assert isinstance(element, (VectorElement, TensorElement))

            # Determine whether this is a non-scalar subargument. This information is later
            # used by the Indexed-Mapper to make some indices loop indices!
            self.argshape = len(element.value_shape())

            # If this is a vector element, we need add an additional accumulation loop iname
            from dune.perftool.pdelab.geometry import dimension_iname
            for i in range(self.argshape):
                self.inames.append(dimension_iname(context='arg', count=i))

            # For the purpose of basis evaluation, we need to take the leaf element
            leaf_element = element.sub_elements()[0]

        # Have the issued instruction depend on the iname for this localfunction space
        iname = lfs_iname(leaf_element, restriction, o.number())
        self.inames.append(iname)

        if self.grad:
            from dune.perftool.pdelab.argument import name_testfunction_gradient
            return Subscript(Variable(name_testfunction_gradient(leaf_element, restriction)), (Variable(iname),))
        else:
            from dune.perftool.pdelab.argument import name_testfunction
            return Subscript(Variable(name_testfunction(leaf_element, restriction)), (Variable(iname),))

    def coefficient(self, o):
        # Do something different for trial function and coefficients from jacobian apply
        if o.count() == 0 or o.count() == 1:
            # Correct the restriction on boundary integrals
            restriction = self.restriction
            if self.measure == 'exterior_facet':
                restriction = Restriction.NEGATIVE

            if self.grad:
                if o.count() == 0:
                    from dune.perftool.pdelab.argument import name_trialfunction_gradient
                    return Variable(name_trialfunction_gradient(o.ufl_element(), restriction, self.component))
                else:
                    from dune.perftool.pdelab.argument import name_apply_function_gradient
                    return Variable(name_apply_function_gradient(o.ufl_element(), restriction, self.component))
            else:
                if o.count() == 0:
                    from dune.perftool.pdelab.argument import name_trialfunction
                    return Variable(name_trialfunction(o.ufl_element(), restriction, self.component))
                else:
                    from dune.perftool.pdelab.argument import name_apply_function
                    return Variable(name_apply_function(o.ufl_element(), restriction, self.component))

        # Check if this is a parameter function
        else:
            # We expect all coefficients to be of type Expression or VectorExpression!
            from dune.perftool.ufl.execution import Expression, VectorExpression
            assert isinstance(o, (Expression, VectorExpression))

            # Determine the name of the parameter function
            from dune.perftool.generation import get_global_context_value
            name = get_global_context_value("data").object_names[id(o)]

            from ufl.checks import is_cellwise_constant
            cellwise_constant = is_cellwise_constant(o)

            # Trigger the generation of code for this thing in the parameter class
            from dune.perftool.pdelab.parameter import (cell_parameter_function,
                                                        intersection_parameter_function,
                                                        )
            if o.on_intersection:
                intersection_parameter_function(name, o, cellwise_constant)
            else:
                cell_parameter_function(name, o, self.restriction, cellwise_constant)

            # And return a symbol
            return Variable(name)

    def indexed(self, o):
        """ Although indexed is something that we want to handle in the pymbolic mapper we
        need to handle it here, as we have to give special treatment to indices into vector
        elements: they become a loop index! """
        # Set the generated index for the argument handler to have it available
        self.last_index = self.call(o.ufl_operands[1])
        aggr = self.call(o.ufl_operands[0])

        use_indices = self.last_index[self.argshape:]

        self.argshape = 0
        if isinstance(aggr, Subscript):
            return Subscript(aggr.aggregate, aggr.index + use_indices)
        else:
            return Subscript(aggr, use_indices)

    def index_sum(self, o):
        # There is three scenarios here:
        # * This is a normal IndexSum:
        #   We should collect the reduction inames, go into recursion and issue
        #   a reduction operation to a temporary
        # * This is part of a nested indexsum:
        #   We only collect the reduction iname and go into recursion
        # * This IndexSum is implicitly handled by the accumulation process
        #   We only go into recursion

        # Get the iname for the reduction index
        ind = o.ufl_operands[1][0]
        self.redinames = self.redinames + (ind,)
        shape = o.ufl_operands[0].ufl_index_dimensions[0]
        from dune.perftool.pdelab import name_index
        domain(name_index(ind), shape)

        # If the left operand is an index sum to, we do it in one reduction
        from ufl.classes import IndexSum
        if isinstance(o.ufl_operands[0], IndexSum):
            return self.call(o.ufl_operands[0])
        else:
            from loopy import Reduction

            # Recurse to get the summation expression
            term = self.call(o.ufl_operands[0])
            self.redinames = tuple(i for i in self.redinames if i not in self.dimension_indices)
            if len(self.redinames) > 0:
                ret = Reduction("sum", tuple(name_index(ind) for ind in self.redinames), term)
                name = get_temporary_name()
                # Generate a substitution rule for this one.
                from loopy import SubstitutionRule
                self.substitution_rules.append(SubstitutionRule(name,
                                                                (),
                                                                ret
                                                                )
                                               )

                ret = Variable(name)
            else:
                ret = term

            # Reset the reduction inames for future indexsums
            self.redinames = ()

            return ret

    def _index_or_fixed_index(self, index):
        from ufl.classes import FixedIndex
        if isinstance(index, FixedIndex):
            return index._value
        else:
            from pymbolic.primitives import Variable
            from dune.perftool.pdelab import name_index
            if index in self.dimension_indices:
                from dune.perftool.pdelab.geometry import dimension_iname
                self.inames.append(self.dimension_indices[index])
                return Variable(self.dimension_indices[index])
            else:
                return Variable(name_index(index))

    def multi_index(self, o):
        return tuple(self._index_or_fixed_index(i) for i in o)

    def index(self, o):
        return self._index_or_fixed_index(o)
